#!/usr/bin/env python
# coding: utf-8

import argparse
import hashlib
import re
import subprocess
import uuid

dtypes = {'int': 'int', 'unsignedInt': 'unsignedInt', 'hexBinary': 'hexBinary', 'NOTATION': 'NOTATION',
          'nonPositiveInteger': 'nonPositiveInteger', 'float': 'float', 'ENTITY': 'ENTITY', 'bool': 'boolean',
          'positiveInteger': 'positiveInteger', 'duration': 'duration', 'IDREFS': 'IDREFS',
          'unsignedLong': 'unsignedLong', 'normalizedString': 'normalizedString', 'dateTimeStamp': 'dateTimeStamp',
          'NMTOKEN': 'NMTOKEN', 'negativeInteger': 'negativeInteger', 'base64Binary': 'base64Binary',
          'long': 'long', 'unsignedShort': 'unsignedShort', 'ENTITIES': 'ENTITIES', 'anyURI': 'anyURI',
          'NMTOKENS': 'NMTOKENS', 'IDREF': 'IDREF', 'unsignedByte': 'unsignedByte', 'Name': 'Name',
          'dayTimeDuration': 'dayTimeDuration', 'date': 'date', 'integer': 'integer', 'byte': 'byte', 'ID': 'ID',
          'gMonth': 'gMonth', 'short': 'short', 'language': 'language', 'gMonthDay': 'gMonthDay',
          'double': 'double', 'Decimal': 'decimal', 'gDay': 'gDay', 'gYearMonth': 'gYearMonth',
          'QName': 'QName', 'datetime': 'dateTime', 'nonNegativeInteger': 'nonNegativeInteger', 'gYear': 'gYear',
          'token': 'token', 'time': 'time', 'yearMonthDuration': 'yearMonthDuration', 'NCName': 'NCName',
          'str': 'string'}


def find_all_locs(string, substring):
    locs = [m.start() for m in re.finditer(substring, string)]
    return locs


def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def find_identifier(entity_name, prov_file):
    hash_value = None

    if entity_name[:4] != 'run:':
        try:
            hash_value = md5(entity_name)
        except:
            print('entity_name is un-hashable')

    ID = []

    with open(prov_file, 'r') as f:
        for line in f:
            if hash_value is not None:
                is_there_hash = line.find(hash_value)
                if is_there_hash != -1:
                    loc1 = line.find('(')
                    loc2 = line.find(',')
                    ID.append(line[loc1 + 1:loc2])
            entity_loc = line.find(entity_name)
            if entity_loc != -1:
                loc1 = line.find('(')
                loc2 = line.find(',')
                ID.append(line[loc1 + 1:loc2])

    id_set = list(set(ID))

    return id_set


def find_activity_id(prov_file_name):
    with open(prov_file_name, 'r') as f:
        for line in f:
            if line[:22] == 'agent(urn_uuid:praetor':
                agent_id = line[line.find('(') + 1:line.find(',')]
                return agent_id


def escape_forbidden_characters(in_string):
    forbidden = "\\(=)|,:;[]"
    allowed_special = '/._-'
    no_forbidden = [x if x.isalnum() else '\\' + x if x in forbidden else x if x in allowed_special else '' for x in
                    in_string]
    final_string = ''.join(no_forbidden)
    return final_string


def add_quality_prov_prtr(input_prov_file, prov_ID, obj_type, name, value, quality_file='quality.provn',
                          temp_listing='listing.txt', preserve=None, other_prov=None, other_prov_id=None):
    ID = 'run:' + str(uuid.uuid4())
    val_type = type(value).__name__
    prov_val_type = dtypes[val_type]

    if not preserve:
        name = escape_forbidden_characters(name)
        value = escape_forbidden_characters(value)

    Q_ID = 'run:quality_measure_' + name

    prefixes = '''
    prefix urn_uuid <urn:uuid:>
    prefix run <http://example.org/>
    prefix var <http://openprovenance.org/var#>
    prefix rdf <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    prefix rdfs <http://www.w3.org/2000/01/rdf-schema#>
    prefix prtr <https://praetor.pages.mpcdf.de/praetor_provenance/>
    '''
    qm_string = 'entity({0}, [prov:value="{1}" %% xsd:{2}, rdf:type="{3}"])'.format(ID, value, prov_val_type, Q_ID)
    quality_label = 'entity({0}, [prov:label="{1}", rdfs:subClassOf="prtr:QualityMeasure"])'.format(Q_ID, name)
    if obj_type == 'entity':
        quality_entity = 'entity({0},[prtr:hadQuality="{1}"])'.format(prov_ID, ID)
    elif obj_type == 'activity':
        quality_entity = 'activity({0},[prtr:hadQuality="{1}"])'.format(prov_ID, ID)
    elif obj_type == 'agent':
        quality_entity = 'agent({0},[prtr:hadQuality="{1}"])'.format(prov_ID, ID)
    else:
        raise ValueError(
            'Allowed values for the type of object to attach quality to are (--type, -t) are entity, activity, or agent')

    if other_prov is None:
        prov_file_string = 'document\n{}\n{}\n{}\n{}\nendDocument'.format(prefixes, qm_string, quality_label,
                                                                          quality_entity)
    else:
        if other_prov_id is None:
            other_prov_id = find_activity_id(other_prov)
        file_relations = 'wasGeneratedBy({0},{1},-)'.format(ID, other_prov_id)
        prov_file_string = 'document\n{0}\n{1}\n{2}\n{3}\n{4}\nendDocument'.format(prefixes, qm_string,
                                                                                   quality_label, quality_entity,
                                                                                   file_relations)

    with open(quality_file, 'w') as f:
        f.write(prov_file_string)

    listing_text = 'file, {0}, provn\nfile, ./{1}, provn'.format(input_prov_file, quality_file)

    with open(temp_listing, 'w') as f:
        f.write(listing_text)

    outfile = input_prov_file.replace('.provn', '_quality.provn')

    subprocess.call(['provconvert', '-merge', temp_listing, '-outfile', outfile])

    if other_prov is not None:
        listing_text = 'file, {0}, provn\nfile, ./{1}, provn'.format(outfile, other_prov)
        with open(temp_listing, 'w') as f:
            f.write(listing_text)
        subprocess.call(['provconvert', '-merge', temp_listing, '-outfile', outfile])

if __name__ == "__main__":

    parser = argparse.ArgumentParser('My program')
    parser.add_argument('-f', '--provfile')
    parser.add_argument('-i', '--provid')
    parser.add_argument('-q', '--quality')
    parser.add_argument('-v', '--value')
    parser.add_argument('-e', '--entity')
    parser.add_argument('-p', '--preserve')
    parser.add_argument('-w', '--workflow')
    parser.add_argument('-t', '--type')

    args = parser.parse_args()

    prov_file = args.provfile
    prov_id = args.provid
    quality_name = args.quality
    quality_value = args.value
    object_value = args.entity
    preserve_gate = args.preserve
    creation_prov_file = args.workflow
    object_type = args.type

    if prov_id is None and object_value is None:
        lifeline_ID = find_activity_id(prov_file)
        object_type = 'agent'
        add_quality_prov_prtr(prov_file, lifeline_ID, object_type, quality_name, quality_value, preserve=preserve_gate,
                              other_prov=creation_prov_file)
    elif object_value == None:
        add_quality_prov_prtr(prov_file, prov_id, object_type, quality_name, quality_value, preserve=preserve_gate,
                              other_prov=creation_prov_file)
    elif prov_id == None:
        ID = find_identifier(object_value, prov_file)
        if len(ID) == 0:
            print('specified value and hash value not found in provenance')
        else:
            object_type = 'entity'
            for identifier in ID:
                add_quality_prov_prtr(prov_file, identifier, object_type, quality_name, quality_value,
                                      preserve=preserve_gate, other_prov=creation_prov_file)
    else:
        print('Please use either a prov ID or entity, not both')



#!/usr/bin/env python
# coding: utf-8

from collections import Counter
import datetime
import hashlib
import inspect
import json
import os
import re
import subprocess
import sys
import time
import types
import uuid



try:
    prtr_path = os.environ['PRAETOR']
    sys.path.append(prtr_path)
    import praetor_settings_user as praetor_settings
except:
    from praetor import praetor_settings

sys.path.append(praetor_settings.provenance_directory + os.environ['PRAETOR_pipeline_id'] + '/')

from praetor import praetorTemplates

import templateDictionary

if praetor_settings.memory:
    from memory_profiler import memory_usage

ts = time.time()

bindings = {
    "context": {"xsd": "http://www.w3.org/2001/XMLSchema#", "lsim": "http://example.org/", "urn_uuid": "urn:uuid:",
                "run": "http://example.org/"}}

session_id = 'praetor'
prov_id_counter = Counter()
prov_id_cache = dict()
prov_id_activity_cache = []
call_func_list = []
glob_vars = [{}]


dtypes = {'int': 'int', 'unsignedInt': 'unsignedInt', 'hexBinary': 'hexBinary', 'NOTATION': 'NOTATION',
          'nonPositiveInteger': 'nonPositiveInteger', 'float': 'float', 'ENTITY': 'ENTITY', 'bool': 'boolean',
          'positiveInteger': 'positiveInteger', 'duration': 'duration', 'IDREFS': 'IDREFS',
          'unsignedLong': 'unsignedLong', 'normalizedString': 'normalizedString', 'dateTimeStamp': 'dateTimeStamp',
          'NMTOKEN': 'NMTOKEN', 'negativeInteger': 'negativeInteger', 'base64Binary': 'base64Binary',
          'long': 'long', 'unsignedShort': 'unsignedShort', 'ENTITIES': 'ENTITIES', 'anyURI': 'anyURI',
          'NMTOKENS': 'NMTOKENS', 'IDREF': 'IDREF', 'unsignedByte': 'unsignedByte', 'Name': 'Name',
          'dayTimeDuration': 'dayTimeDuration', 'date': 'date', 'integer': 'integer', 'byte': 'byte', 'ID': 'ID',
          'gMonth': 'gMonth', 'short': 'short', 'language': 'language', 'gMonthDay': 'gMonthDay',
          'double': 'double', 'Decimal': 'decimal', 'gDay': 'gDay', 'gYearMonth': 'gYearMonth',
          'QName': 'QName', 'datetime': 'dateTime', 'nonNegativeInteger': 'nonNegativeInteger', 'gYear': 'gYear',
          'token': 'token', 'time': 'time', 'yearMonthDuration': 'yearMonthDuration', 'NCName': 'NCName',
          'str': 'string'}



def md5(fname):
    '''
    Function to create hash codes that represent data, files, etc.
    :param fname: Name of file to read and hash
    :return: Hash code specific to file fname
    '''
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def get_agent_string(modules):
    '''
    Create unique id for the agent of a pipeline and get the version of python and all imported modules
    :param modules: List of module names imported
    :return: Unique id for the agent and list of versions for modules
    '''
    versions = []
    procPy = subprocess.Popen(['python', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    pythonVersion = procPy.stderr.read()

    loc = 'Python '
    pyVer = pythonVersion[pythonVersion.find(loc)+len(loc):]
    pyVer = pyVer.strip()
    pyVer = ''.join(e for e in pyVer if e.isalnum() or e == '.' or e == '-')
    pyVer = 'run:'+pyVer

    for mod in modules:
        proc1 = subprocess.Popen(['pip', 'show', mod], stdout=subprocess.PIPE)
        proc2 = subprocess.Popen(['grep', 'Version'], stdin=proc1.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc1.stdout.close()

        out, err = proc2.communicate()
        versions.append(out)

    versions = [x.replace('\n', '') for x in versions]
    versions = [x[x.find(': ')+2:] for x in versions]

    modString = [x + '_' + y for x, y in zip(modules, versions)]
    modString = ', '.join(modString)

    return modString, pyVer


def get_modules():
    '''
    Function to create the agent_json.json file, including creating a unique id for the agent, determining all modules
    imported, find their versions, structure and input all infroamtion into agent_json.json
    :return: agent_json.json
    '''
    modules = []
    for name, val in globals().items():
        if isinstance(val, types.ModuleType):
            modName = val.__name__
            loc = modName.find('.')
            if loc != -1:
                modName = val.__name__[0:loc]
            modules.append(modName)

    exclude = ['__builtin__',
               'Tester','_thread','graphlib','scriptScript','__future__','_threading_local','grp','scriptresubmitter',
               '_abc','_tkinter','gzip','secrets','_aix_support','_tracemalloc','hashlib','select','_ast','_uuid',
               'heapq','selectors','_asyncio','_warnings','hmac','setuptools','_bisect','_weakref','html','shelve',
               '_blake2','_weakrefset','http','shlex','_bootlocale','_xxsubinterpreters','idlelib','shutil',
               '_bootsubprocess','_xxtestfuzz','imaplib','signal','_bz2','_zoneinfo','imghdr','site','_codecs','abc',
               'imp','smtpd','_codecs_cn','aifc','importlib','smtplib','_codecs_hk','antigravity','inspect','sndhdr',
               '_codecs_iso2022','argparse','io','socket','_codecs_jp','array','ipaddress','socketserver','_codecs_kr',
               'ast','itertools','spwd','_codecs_tw','asynchat','json','sqlite3','_collections','asyncio','keyword',
               'sre_compile','_collections_abc','asyncore','latexWriter','sre_constants','_compat_pickle','atexit',
               'lib2to3','sre_parse','_compression','audioop','linecache','ssl','_contextvars','base64','locale',
               'stat','_crypt','bdb','logging','statistics','_csv','binascii','lzma','string','_ctypes','binhex',
               'mailbox','stringprep','_ctypes_test','bisect','mailcap','struct','_curses','builtins','marshal',
               'subprocess','_curses_panel','bwd','math','sunau','_datetime','bz2','mimetypes','symbol','_decimal',
               'cProfile','mmap','symtable','_distutils_hack','calendar','modulefinder','sys','_elementtree','cgi',
               'multiprocessing','sysconfig','_functools','cgitb','netrc','syslog','_hashlib','chunk','nis','tabnanny',
               '_heapq','cmath','nntplib','tarfile','_imp','cmd','ntpath','telnetlib','_io','code','nturl2path',
               'tempfile','_json','codecs','numbers','termios','_locale','codeop','opcode','test','_lsprof',
               'collections','operator','tester','_lzma','colorsys','optparse','testing_prtr_packagetemplateDictionary',
               '_markupbase','compileall','os','textwrap','_md5','concurrent','ossaudiodev','this','_multibytecodec',
               'configparser','parser','threading','_multiprocessing','contextlib','pathlib','time','_opcode',
               'contextvars','pdb','timeit','_operator','copy','periodic','tkinter','_osx_support','copyreg','pickle',
               'token','_peg_parser','crypt','pickletools','tokenize','_pickle','csv','pip','trace','_posixshmem',
               'ctypes','pipes','traceback','_posixsubprocess','curses','pkg_resources','tracemalloc','_py_abc',
               'dataclasses','pkgutil','tty','_pydecimal','datetime','platform','turtle','_pyio','dbm','plistlib',
               'turtledemo','_queue','decimal','poplib','types','_random','difflib','posix','typing','_sha1','dis',
               'posixpath','unicodedata','_sha256','distutils','pprint','unittest','_sha3','doctest','profile','urllib',
               '_sha512','email','provcall-NoWorkflow-save','uu','_signal','encodings','pstats','uuid','_sitebuiltins',
               'ensurepip','pty','valuefilemaker','_socket','enum','pwd','venv','_sqlite3','errno','py_compile',
               'warnings','_sre','faulthandler','pyclbr','wave','_ssl','fcntl','pydoc','weakref','_stat','filecmp',
               'pydoc_data','webbrowser','_statistics','fileinput','pyexpat','wheel','_string','fnmatch','pytest',
               'wsgiref','_strptime','formatter','python_modules','xdrlib','_struct','fractions','queue','xml',
               '_symtable','ftplib','quopri','xmlrpc','_sysconfigdata__linux_x86_64-linux-gnu','functools',
               'random','xxlimited','_sysconfigdata_x86_64_conda_cos6_linux_gnu','gc','re','xxsubtype',
               '_sysconfigdata_x86_64_conda_linux_gnu','genBindings','readline','zerosigmaker','_testbuffer',
               'genericpath','reprlib','zipapp','_testcapi','getopt','resource','zipfile','_testimportmultiple',
               'getpass','rlcompleter','zipimport','_testinternalcapi','gettext','runpy','zlib','_testmultiphase',
               'glob','sched','zoneinfo',
               'genBindings','praetor','convert2prov','decorate','praetorTemplates','praetor_settings',
               'praetor_settings_user', 'templateDictionary', 'python_modules']

    modules = [x for x in modules if x not in exclude]

    json_total = {}
    modules_versions, py_version = get_agent_string(modules)

    bindings['var'] = {}
    bindings['var']['modules'] = ['run:({})'. format(modules_versions)]
    bindings['var']['python_version'] = [py_version]
    bindings['var']['lifeline'] = ['urn_uuid:{}'.format(os.environ["PRAETOR_pipeline_id"])]

    json_dir = praetor_settings.provenance_directory + os.environ['PRAETOR_pipeline_id'] + '/json/'
    json_total['agent'] = bindings

    with open(json_dir + 'agent_json.json', 'w') as f:
        json.dump(json_total, f)

def timestamp():
    ts = str(time.time())
    return ts


def datetimestamp():
    dts = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]
    return dts


def gen_identifier(variable, naming_template="entity"):
    '''
    General utility function to create prov ids for python objects
    :param variable: Value of object to create id for
    :param naming_template: Name related to the object in question to include in the id
    :return: id for target object
    '''
    try:
        prov_id = prov_id_cache[id(variable)]
        if 'execute_inEntity_0_praetor_5' in prov_id:
            write_string = 'provid - ' + prov_id + '\nvalue - ' + variable + '\npythonid - ' + str(id(variable)) + '\n'
            with open(praetor_settings.provenance_directory + 'send_me_this_file_kristen.txt', 'a') as f:
                f.write(write_string)


    except KeyError:
        prov_id_counter[naming_template] += 1
        prov_id = '{}_{}_{}'.format(naming_template, session_id, prov_id_counter[naming_template])
        prov_id_cache[id(variable)] = prov_id

    return prov_id


def escape_forbidden_characters(in_string):
    '''
    Escape forbidden characters in input string in accordance with PROV-N
    :param in_string: input string to escape characters within
    :return: modified in_string with escaped characters
    '''
    forbidden = "\\(=)|,:;[]"
    allowed_special = '/._-'
    no_forbidden = [x if x.isalnum() else '\\' + x if x in forbidden else x if x in allowed_special else '' for x in
                 in_string]
    final_string = ''.join(no_forbidden)
    return final_string


def remove_quotes_from_string(in_string):
    out_string = [x for x in in_string if x not in ['"', "'"]]
    out_string = ''.join(out_string)
    return out_string


def find_type(value):
    '''
    If possible, return the type of input value and convert it to string
    :param value: Input value to type
    :return: String value, type of value
    '''
    try:
        outVal = str(value)
        outVal = remove_quotes_from_string(outVal)
        otype = type(value).__name__
        if otype in dtypes.keys():
            ptype = dtypes[otype]
        else:
            ptype = 'string'
    except:
        outVal = str(value)
        outVal = remove_quotes_from_string(outVal)
        ptype = 'string'
    return outVal, ptype


def reduce_lists(in_list):
    '''
    Sub-rountine for un-nesting lists if they are present
    :param in_list: Nested input list
    :return: Flattened version of in_list
    '''
    while any(isinstance(i, list) for i in in_list):
        flat_new = []
        for sublist in in_list:
            if not isinstance(sublist, list):
                flat_new.append(sublist)
            else:
                for value in sublist:
                    flat_new.append(value)
        in_list = flat_new

    return in_list


def extract_wrapped(decorated):
    '''
    Extract the source code for a wrapped function
    :param decorated: Function source code including decorator
    :return: Function source code excluding decorator
    '''
    closure = (c.cell_contents for c in decorated.__closure__)
    return next((c for c in closure if isinstance(c, types.FunctionType)), None)



def find_dict_differences(dict1, dict2):
    '''
    Used to find the differences between two dictionaries of global variables
    :param dict1: initial global variable dictionary
    :param dict2: updated global variable dictionary
    :return: Items in dict2 that are not in dict1
    '''
    differences = {}

    for key in dict2.keys():
        if key not in dict1:
            differences[key] = (dict2[key])

    for key in set(dict1.keys()).intersection(dict2.keys()):
        try:
            if dict1[key] != dict2[key]:
                differences[key] = ( dict2[key])
        except ValueError:
            if str(dict1[key]) != str(dict2[key]):
                differences[key] = (dict2[key])

    return differences


def track_globs(initial_globals):
    '''
    Find changes in global variables between and old set and a freshly generated one
    :param initial_globals: Set of global variables to compare to
    :return: Differences between current and previous global variables
    '''
    final_globals = dict(globals())
    changed_globals = {name: value for name, value in final_globals.items() if name not in initial_globals or initial_globals[name] != value}
    return changed_globals


def provcall(inlist, outlist, template, longname, startTime, endTime, memUse, json_name,
             json_out='full_json.json', jsonDir='./', maxValueLength=praetor_settings.max_value_length, module_name='main',
             global_inputs=None, global_outputs=None):

    '''
    Consumes as input the various parameters and metrics surrounding a function execution and writes them into a json
    file as bindings, ready to combine with a template and make provenance.
    :param inlist: List of input values
    :param outlist: List of output values
    :param template: Name of provenance template corresponding to the function
    :param longname: Full name of the function (String)
    :param startTime: Start time of execution (dateTime object)
    :param endTime: End time of execution (dateTime object)
    :param memUse: Average memory consumption during function (Bytes)
    :param json_name: String for key to identify function execution in json file
    :param json_out: Name of json file to write data to
    :param jsonDir: Directory where json_out is located
    :param maxValueLength: Max number of characters for an input/output before it is saved as a "big entity"
    :param module_name: Name of thing where the function is defined (either main or a module)
    :param global_inputs: Dictionary of global variables consumed by a function
    :param global_outputs: Dictionary of global variables produced by a function
    :return: Updates json_out with the above information for the target function execution
    '''
    template = ''.join(e if e.isalnum() else e if e == '_' else '' for e in template)
    longname = ''.join(e if e.isalnum() else e if e == '_' else '' for e in longname)
    identifier = uuid.uuid4()
    json_total = {}
    bindings['var'] = {}
    bindings['var']['messageStartTime'] = [{"@type": "xsd:dateTime", "@value": startTime}]
    bindings['var']['moduleName'] = [{"@type": "xsd:string", "@value": module_name}]
    bindings['var']['activityName'] = [{"@type": "xsd:string", "@value": longname}]

    if len(call_func_list) != 0 and call_func_list[-1] != '<module>':
        bindings['var']['message2'] = [{"@id": prov_id_activity_cache[-1][0]}]
        bindings['var']['message2StartTime'] = [{"@type": "xsd:dateTime", "@value": prov_id_activity_cache[-1][1]}]
        bindings['var']['message2EndTime'] = [{"@type": "xsd:dateTime", "@value": prov_id_activity_cache[-1][2]}]
        bindings['var']['template'] = [template + 'started']
    else:
        bindings['var']['template'] = [template]
    bindings['var']['message'] = [{"@id": "urn_uuid:{}_{}".format(longname, identifier)}]
    prov_id_activity_cache.append(["urn_uuid:{}_{}".format(longname, identifier),startTime,endTime])

    if longname in templateDictionary.tempDict:
        if len(outlist) < len(templateDictionary.tempDict[longname][1]):
            print('mismatch lengths', outlist)
            lenDiff = len(templateDictionary.tempDict[longname][1]) - len(outlist)
            for extra in range(lenDiff):
                outlist.append('Not Computed')
        if len(inlist) < len(templateDictionary.tempDict[longname][0]):
            print('mismatch input', inlist)
            lenDiff = len(templateDictionary.tempDict[longname][0]) - len(inlist)
            for extra in range(lenDiff):
                inlist.append('Not Computed')

    longfileDir = jsonDir.replace('json', 'big_entities')

    for x in range(len(outlist)):
        if callable(outlist[x]):
            outVal, ptype = find_type('callable_{}'.format(type(outlist[x])))
            object_id = gen_identifier(outVal, naming_template='{}_outEntity_{}'.format(longname, x))
        else:
            outVal, ptype = find_type(outlist[x])
            object_id = gen_identifier(outlist[x], naming_template='{}_outEntity_{}'.format(longname, x))
        if len(outVal) > maxValueLength:
            with open(longfileDir + object_id + '.txt', 'w') as f:
                f.write(str(outlist[x]))
            outVal = longfileDir + object_id + '.txt'
            ptype = 'string'

        bindings['var']['output' + str(x)] = [{"@id": "run:" + object_id}]
        bindings['var']['output' + str(x) + 'value'] = [{"@value": outVal, "@type": "xsd:{}".format(ptype)}]

    if global_outputs is not None:
        object_id = gen_identifier(global_outputs, naming_template='globals')
        bindings['var']['praetorglobaloutput'] = [{"@id": "run:" + object_id}]
        for glob_key, glob_val in global_outputs.iteritems():
            object_id = gen_identifier(glob_val, naming_template='globals_{}'.format(glob_key))
            g_val, ptype = find_type(glob_val)
            if len(g_val) > maxValueLength:
                with open(longfileDir + object_id + '.txt', 'w') as f:
                    f.write(str(glob_val))
                g_val = longfileDir + object_id + '.txt'
                ptype = 'string'
            bindings['var']['praetorglobaloutput{}'.format(glob_key)] = [{"@id": "run:" + object_id,
                                                                          "@value": g_val,
                                                                          "@type": "xsd:{}".format(ptype),
                                                                          "@label": 'run:{}'.format(glob_key)}]

    for x in range(len(inlist)):
        if callable(inlist[x]):
            inVal, ptype = find_type('callable_{}'.format(type(inlist[x])))
            object_id = gen_identifier(inVal, naming_template='{}_inEntity_{}'.format(longname, x))
        else:
            inVal, ptype = find_type(inlist[x])
            object_id = gen_identifier(inlist[x], naming_template='{}_inEntity_{}'.format(longname, x))
        if len(inVal) > maxValueLength:
            with open(longfileDir + object_id + '.txt', 'w') as f:
                f.write(str(inlist[x]))
            inVal = longfileDir + object_id + '.txt'
            ptype = 'string'
        bindings['var']['input' + str(x)] = [{"@id": "run:" + object_id}]
        bindings['var']['input' + str(x) + 'value'] = [{"@value": inVal, "@type": "xsd:{}".format(ptype)}]

    if global_inputs is not None:
        object_id = gen_identifier(global_inputs, naming_template='globals')
        bindings['var']['praetorglobalinput'] = [{"@id": "run:" + object_id}]
        for glob_key, glob_val in global_inputs.iteritems():
            object_id = gen_identifier(glob_val, naming_template='globals_{}'.format(glob_key))
            g_val, ptype = find_type(glob_val)
            if len(g_val) > maxValueLength:
                with open(longfileDir + object_id + '.txt', 'w') as f:
                    f.write(str(glob_val))
                g_val = longfileDir + object_id + '.txt'
                ptype = 'string'
            bindings['var']['praetorglobalinput{}'.format(glob_key)] = [{"@id": "run:" + object_id, "@value": g_val,
                                                                         "@type": "xsd:{}".format(ptype),
                                                                         "@label": 'run:{}'.format(glob_key)}]
    bindings['var']['messageEndTime'] = [{"@type": "xsd:dateTime", "@value": endTime}]
    bindings['var']['memoryUsage'] = [{"@type": "xsd:string", "@value": memUse}]


    if len(call_func_list) != 0 and call_func_list[-1] != '<module>':
        json_total[json_name + 'started'] = bindings
        with open(jsonDir + json_out, 'a') as f:
            json.dump(json_total, f)
            f.write('\n')

    else:
        json_total[json_name] = bindings
        with open(jsonDir + json_out, 'a') as f:
            json.dump(json_total, f)
            f.write('\n')


def printDict():
    print(prov_id_cache, "prov_id_cache")


def for_all_methods(decorator):
    '''
    Applies a target decorator to a class, was used to apply the provWrap but is currently out of use
    :param decorator: Decorator to add to a class
    :return: Class with altered functionality to record provenance
    '''
    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)):
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls

    return decorate


def get_globals():
    '''
    Creates a dictionary of all global variables present within the target pipeline, execluding those added by the
    praetor  code and built-in variables
    :return: Dictionary of global variables
    '''
    frame_1 = inspect.currentframe().f_back
    frame = frame_1.f_back
    globals_dict = frame.f_globals
    main_script_globals = {key:value for key,value in globals_dict.items() if not key.startswith('__')
                           and not any(sub in key for sub in ['genBindings', 'praetor', 'prov_id_counter'])
                           and not (inspect.ismodule(value) or inspect.isfunction(value) or inspect.isclass(value))}
    return main_script_globals


def create_template(func, template_name, template_name_started, function_name, filePath, global_tracking=False):
    '''
    Create a PROV-N template for a target python function
    :param func: Target function object
    :param template_name: Name of PROV-N template to create
    :param template_name_started: Name of PROV-N template if the function was started by another
    :param function_name: Name of function
    :param filePath: Path where templates will be generated
    :param global_tracking: Whether to track global variables
    :return: Creates the appropriate template for the function and stores a copy of the source code
    '''
    if isinstance(func, types.FunctionType):
        try:
            func = extract_wrapped(func)
        except (TypeError, AttributeError) as e:
            pass

        source = inspect.getsource(func)
        source_str = source.encode('ascii', 'ignore')
        output_list = praetorTemplates.separate_output(source_str)
        arg_names = (inspect.getargspec(func)[0])
        input_list = reduce_lists(arg_names)
        funct_name = func.__name__
        function_parameters = [funct_name, input_list, output_list]
        temp_sep_dict = {function_parameters[0]: [function_parameters[1], function_parameters[2]]}
        reload(templateDictionary)
        full_temp_dict = templateDictionary.tempDict
        full_temp_dict.update(temp_sep_dict)

        bundle, bundleStarted = praetorTemplates.generateBundle(function_parameters, praetorTemplates.bundleStart,
                                                                praetorTemplates.bundleEnd)

        with open(filePath + 'templates/' + template_name, 'w') as f:
            f.write(bundle)

        with open(filePath + 'templates/' + template_name_started, 'w') as f:
            f.write(bundleStarted)

        with open(filePath + 'function_store/' + function_name, 'w') as f:
            f.write(source_str)

        with open(filePath + 'templateDictionary.py', 'w') as f:
            f.write('tempDict = {}'.format(full_temp_dict))



def provWrap(func):
    '''
    Wrapper to decorate a function with in order to record provenance
    :param func: Target function object to wrap
    :return: An altered version of func which records provenance
    '''
    def wrap(*args, **kwargs):
        name = func.__name__
        if name in praetor_settings.blacklist:
            return func(*args, **kwargs)
        try:
            module_loc = func.__module__
            if module_loc == '__main__':
                module_loc = 'main'
                module_name = ''
            else:
                module_name = '_' + module_loc
        except Exception:
            module_loc = 'undefined'
            module_name = ''

        longname = name
        template = name + module_name + '_template.provn'
        json_name = name + module_name
        template_name_started = name + module_name + 'started_template.provn'
        function_name = name + module_name + 'function.txt'
        inlist = list(args)

        filePath = praetor_settings.provenance_directory + os.environ['PRAETOR_pipeline_id'] + '/'
        jsonDir = filePath + 'json/'

        if isinstance(func, types.FunctionType):
            arg_names = (inspect.getargspec(func)[0])
            input_list = reduce_lists(arg_names)
        else:
            input_list = []

        if not os.path.isfile(filePath + 'templates/' + template):
            create_template(func, template, template_name_started, function_name, filePath)

        try:
            a = inspect.getargspec(func)
            defaultsKeys = a.args[-len(a.defaults):]
            defaultsVals = list(a.defaults)
        except TypeError:
            defaultsKeys = []
            defaultsVals = []

        kwargList = []
        for key in input_list:
            if key in kwargs.keys():
                kwargList.append(kwargs[key])
            elif key in defaultsKeys:
                kwargList.append(defaultsVals[defaultsKeys.index(key)])

        inlist.extend(kwargList)


        if not praetor_settings.global_tracking:
            global_inputs = None
        else:
            global_start = get_globals()
            global_inputs = find_dict_differences(glob_vars[0], global_start)
            glob_vars.insert(0, global_start)


        startTime = datetimestamp()
        print('{} function_start: {}.{}'.format(praetor_settings.unique_string, module_loc, name))
        if praetor_settings.memory:
            try:
                mem_use, output = memory_usage((func, args, kwargs), retval=True)
                max_mem = str(max(mem_use))
            except:
                output = func(*args, **kwargs)
                max_mem = '0'
        else:
            output = func(*args, **kwargs)
            max_mem = '0'

        if not praetor_settings.global_tracking:
            global_outputs = None
        else:
            global_variables = get_globals()
            global_outputs = find_dict_differences(glob_vars[0], global_variables)
            glob_vars.insert(0, global_variables)

        print('{} function_end: {}.{}'.format(praetor_settings.unique_string, module_loc, name))
        endTime = datetimestamp()
        outlist = [output]

        provcall(inlist, outlist, template, longname, startTime, endTime, max_mem, json_name, jsonDir=jsonDir,
                 module_name=module_loc, global_outputs=global_outputs, global_inputs=global_inputs)
        call_func_list.append(inspect.stack()[1][3])
        return output

    return wrap


def provWrapOpen(func):
    '''
    Wrapper alter pythons built-in open function so that it records provenance
    :param func: The open function object (typically just open)
    :return: Altered functionality open which also records provenance
    '''
    def wrapOpen(*args, **kwargs):
        filePath = praetor_settings.provenance_directory + os.environ['PRAETOR_pipeline_id'] + '/'
        provDests = ['/json/', '/prov/', '/templates/', '/big_entities/', '/function_store/']
        provPaths = [filePath + x for x in provDests]
        fileName = args[0]

        try:
            gate = [True for x in provPaths if x in fileName]
        except TypeError:
            gate = [False]

        if any(gate):
            output = func(*args, **kwargs)

        else:
            longname = 'pythonBuiltinFileAccess'
            template = longname + '_template.prov'
            json_name = longname
            jsonDir = filePath + '/json/'
            startTime = datetimestamp()
            if praetor_settings.memory:
                try:
                    mem_use, output = memory_usage((func, args, kwargs), retval=True)
                    max_mem = str(max(mem_use))
                except:
                    output = func(*args, **kwargs)
                    max_mem = '0'
            else:
                output = func(*args, **kwargs)
                max_mem = '0'
            endTime = datetimestamp()
            mode = output.mode
            metaData = os.stat(fileName)
            inlist = [fileName, mode]
            outlist = []
            inlist.extend([x for i, x in enumerate(metaData) if i in [4, 6]])
            hashvalue = md5(args[0])
            inlist.append(hashvalue)
            provcall(inlist, outlist, template, longname, startTime, endTime, max_mem, json_name, jsonDir=jsonDir)
            call_func_list.append(inspect.stack()[1][3])
        return output

    return wrapOpen


def decorate_imported_function(func, decorator):
    '''
    Function to apply a decorator to an imported function
    :param func: Target function object
    :param decorator: Decorator to apply
    :return: Altered function which now records provenance
    '''
    func_type = type(func).__name__
    if isinstance(func, types.FunctionType):
        func = decorator(func)
    elif isinstance(func, types.BuiltinFunctionType):
        func = decorator(func)
    elif isinstance(func, types.ModuleType):
        func = decorate_all_in_module(func, decorator)
    else:
        print('function {} with type {} not supported by praetor and not wrapped'.format(func, func_type))
    return func


def find_all(some_list, to_find):
    return [m.start() for m in re.finditer(to_find, some_list)]


def decorate_all_in_module(module, decorator):
    '''
    Decorate the functions within imported modules
    :param module: Module to decorate
    :param decorator: decorator to apply
    :return: The function of the selected module is altered to record provenance
    '''
    mod_spec_blacklist = [x for x in praetor_settings.blacklist if module.__name__ in x]

    func_blacklist = [x[max(find_all(x,'.'))+1:] for x in mod_spec_blacklist]
    for name in dir(module):
        obj = getattr(module, name)
        try:
            obj_name = obj.__name__
            print('obj_name', obj_name)
        except:
            obj_name = obj
        if obj_name not in func_blacklist:
            if isinstance(obj, types.FunctionType):
                setattr(module, name, decorator(obj))
            elif isinstance(obj, types.BuiltinFunctionType):
                setattr(module, name, decorator(obj))
            elif isinstance(obj_name, types.ModuleType):
                decorate_all_in_module(obj, decorator)
            else:
                pass



#!/usr/bin/env python3
# coding: utf-8

from io import BytesIO
import re
import sys
import tokenize

def findAll(inString,findx):
    '''
    Sub-routine used to find import locations
    :param inString: input string to search
    :param findx: substring to look for
    :return: all start locations of findx
    '''
    return [m.start() for m in re.finditer(findx, inString)]


def find_matching(importLines, modSearch):
    '''
    Find lines which import modules specified within the praetor settings file
    :param importLines: Lines where import substring is present
    :param modSearch: List of modules to search for
    :return: Lines of target file which import the modules in modSearch
    '''
    possible_start_values = [',', ' ']
    possible_end_values = [',', ' ', '\n']

    importLines = [x + '\n' for x in importLines]
    for x in importLines:
        hash_find = [i for i, letter in enumerate(x) if letter == '#']
        if len(hash_find) != 0:
            first_hash = min(hash_find)
            def_find = x.find('def')
            if first_hash < def_find:
                importLines.remove(x)
    period_index = modSearch.find('.')
    if period_index not in [-1, 0]:
        modSearch = modSearch[:modSearch.find('.')]

    mod_variations = []

    for start in possible_start_values:
        for end in possible_end_values:
            mod_variations.append(start + modSearch + end)

    matchingMod = [y for y in importLines if any(mod in y for mod in mod_variations)]
    matchingMod = [x[:-1] for x in matchingMod]
    return matchingMod


def module_string(mod):
    '''
    Creates the code to wrap imported modules and or functions
    :param mod: Module name
    :return: Python code to add to decorated file
    '''
    module = 'genBindings.decorate_all_in_module({},genBindings.provWrap)'.format(mod)
    function = '{0} = genBindings.decorate_imported_function({0},genBindings.provWrap)'.format(mod)

    conditional_string ='''
try:
    if isinstance({0}, types.FunctionType):
        {1}
        print("{3} is function" )
    elif isinstance({0}, types.BuiltinFunctionType):
        {1}
        print("{3} is built-in function" )
    else:
        {2}
        print("{3} is Module" )
except:
    print("Module or function {0} not found")
'''.format(mod, function, module, mod)

    return conditional_string


def extract_commented_code(line, py_ver):
    '''
    Function to extract all code hidden behind comments
    :param line: String line of code
    :return: String of comments code or None if there are none
    '''
    if py_ver == 2:
        try:
            tokens = tokenize.generate_tokens(BytesIO(line.encode('utf-8')).readline)
            if tokens is not None:
                for toktype, tok, start, end, line in tokens:
                    if toktype == tokenize.COMMENT:
                        commented_code = tok[1:].strip()
                        return commented_code
        except tokenize.TokenError:
            return 'None'
    else:
        try:
            tokens = tokenize.tokenize(BytesIO(line.encode('utf-8')).readline)
            if tokens is not None:
                for token in tokens:
                    if token.type == tokenize.COMMENT:
                        commented_code = token.string[1:].strip()
                        return commented_code
        except tokenize.TokenError:
            return 'None'

    return 'None'


def get_indentation(line):
    '''
    Function to return the indentation of any line of python code
    :param line: String of line to check
    :return: Numerical representation of indent
    '''
    stripped_line = line.lstrip()
    indentation = len(line) - len(stripped_line)
    return indentation


def add_fake_functions(exampleScript, loc_key='praetor_insert', loc_key_comment='praetor_comment'):
    '''
    Reads an input script, searches for commented keywords, generates and inserts fake functions into the
    decorated version of the script to record tailored provenance
    :param exampleScript: String of target python script
    :param loc_key: Keyword to trigger fake generation
    :return: original script with fake function additions as a string
    '''
    data = exampleScript.split('\n')
    py_ver = sys.version_info[0]
    to_comment = []
    for i, line in enumerate(data):
        line_comment = extract_commented_code(line, py_ver)
        if loc_key in line_comment:
            line_indent = get_indentation(line)
            com_items = line_comment.split(' ')
            input_names = [x[len('in_names='):] for x in com_items if 'in_names=' in x]
            input_values = [x[len('in_values='):] for x in com_items if 'in_values=' in x]
            outputs = [x[len('out='):] for x in com_items if 'out=' in x]
            name = [x[len('name='):] for x in com_items if 'name=' in x]
            function_string = '{0}def {1}({2}):\n    {0}return {3}\n{1}({4})'.format(' ' * line_indent, name[0],
                                                                                     input_names[0], outputs[0],
                                                                                     input_values[0])
            data[i] += '\n{0}'.format(function_string, ' ' * line_indent)

        if loc_key_comment in line_comment:
            com_items = line_comment.split(' ')
            to_comment = [x[len('comment='):] for x in com_items if 'comment=' in x]

    reformated_script = '\n'.join(data)
    return reformated_script, to_comment


def decorateFile(exampleScript, modules, wrap_open=False):
    '''
    Function to create a decorated version of a python script which wraps the desired imported modules/functions
    and the functions defined within the script in order to record provenance. It also adds imports to required
    modules and inserts fake functions based upon user input comment-hidden commands.
    :param exampleScript: Target script to decorate
    :param modules: List of target modules to decorate
    :param wrap_open: Whether to wrap pythons built-in open function
    :param global_tracking: Whether global variables should be tracked
    :return: Decorated version of exampleScript
    '''
    find1 = '.py'
    loc1 = exampleScript.index(find1)
    scriptShort = exampleScript[0:loc1]

    scriptNew = scriptShort + '_decorated.py'

    with open(exampleScript, 'r') as f:
        ogData = f.read()

    ogData, commented_globals = add_fake_functions(ogData)

    importRequired = '\nfrom praetor import genBindings\nimport types\ngenBindings.get_modules()\n'
    open_wrap = 'open = genBindings.provWrapOpen(open)\n'
    if wrap_open:
        importRequired += open_wrap

    end_of_script = '\ngenBindings.get_modules()'
    first_line_import = ogData[0:ogData.find('\n')]
    if len(first_line_import) != 0:
        if first_line_import[0:7] == 'import ':
            imports = [0]
            imports.extend([x + 1 for x in findAll(ogData, '\nimport ')])
        else:
            imports = [x + 1 for x in findAll(ogData, '\nimport ')]
    else:
        imports = findAll(ogData, '\nimport ')
        imports = [x + 1 for x in imports]
    importLines = [ogData[x:x + ogData[x:].index('\n')] for x in imports]

    if len(first_line_import) != 0:
        if first_line_import[0:5] == 'from ':
            from_imports = [0]
            from_imports.extend([x + 1 for x in findAll(ogData, '\nfrom ')])
        else:
            from_imports = [x + 1 for x in findAll(ogData, '\nfrom ')]
    else:
        from_imports = [x + 1 for x in findAll(ogData, '\nfrom ')]

    from_import_lines = [ogData[x:x + ogData[x:].index('\n')] for x in from_imports]

    ogData_lists = ogData.split('\n')
    first_line = ogData_lists[0]
    shebang_str = '#!/usr/bin/env python3'
    if shebang_str in first_line:
        first_import_loc = ogData.find(ogData_lists[1])
    else:
        first_import_loc = 0
    from_future = ogData.find('from __future__ import')
    future_import = ogData.find('import __future__')
    if from_future != -1 or future_import != -1:
        from_locs = [m.start() for m in re.finditer('from __future__ import', ogData)]
        im_locs = [im.start() for m in re.finditer('import __future__', ogData)]
        im_locs.extend(from_locs)
        final_loc = max(im_locs)
        first_import_loc = ogData[final_loc:].find('\n')

    ogData = ogData[:first_import_loc] + importRequired + ogData[first_import_loc:]

    ogData = ogData + end_of_script

    asString = ' as '
    for mod in modules:
        matchingMod = find_matching(importLines, mod)
        matchingFrom = find_matching(from_import_lines, mod)


        if len(matchingMod) != 0:
            if len(matchingMod) > 1:
                print('two imports matching the same naming convention for lines {} these have not been wrapped'.format(
                    matchingMod))
                continue
            if asString in matchingMod[0]:
                functName = matchingMod[0][matchingMod[0].index(asString) + len(asString):]
            else:
                functName = mod
            modifier = module_string(functName)
            ogData = ogData.replace(matchingMod[0], matchingMod[0] + '\n' + modifier)

        if len(matchingFrom) != 0:
            for line in matchingFrom:
                loc1 = line.find(' import ') + len(' import ')
                if loc1 != -1:
                    new_dec = module_string(mod)
                    ogData = ogData.replace(line, line + '\n' + new_dec)

    updatedData = ogData.replace('\ndef ', '\n@genBindings.provWrap\ndef ')

    with open(scriptNew, 'w') as f:
        f.write(updatedData)

    return scriptNew, commented_globals







#!/usr/bin/env python2
import os
import subprocess
import sys
import uuid
import time

t_start = time.time()

try:
    prtr_path = os.environ['PRAETOR']
    sys.path.append(prtr_path)
    import praetor_settings_user as praetor_settings

except:
    from praetor import praetor_settings
    # import praetor_settings
    pass

def generate_pipeline_id():
    '''
    Creates a unique id for a pipeline run which includes the praetor version
    :return: Unique id
    '''
    proc1 = subprocess.Popen(['pip', 'show', 'praetor'], stdout=subprocess.PIPE)
    proc2 = subprocess.Popen(['grep', 'Version'], stdin=proc1.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    proc1.stdout.close()
    out, err = proc2.communicate()
    out = out.decode("utf-8")
    praetor_version = out[out.find(': ') + 2: -1]

    pipeline_id = "praetor_{}_{}".format(praetor_version, uuid.uuid4())
    return pipeline_id


def escape_forbidden_characters(in_string):
    '''
    Escapes special characters from an input string to fit PROV-N
    :param in_string: String to escape characters within
    :return: Original string with escaped characters
    '''
    forbidden = "\\(=)|,:;[]"
    allowed_special = '/._-'
    no_forbidden = [x if x.isalnum() else '\\' + x if x in forbidden else x if x in allowed_special else '' for x in in_string]
    final_string = ''.join(no_forbidden)
    return final_string


pipeline_id = generate_pipeline_id()
os.environ['PRAETOR_pipeline_id'] = pipeline_id
currentDir = praetor_settings.provenance_directory + '{}/'.format(pipeline_id)
subprocess.call(['mkdir', currentDir])

with open(currentDir+'templateDictionary.py','w') as f:
    f.write('tempDict = {}')


from praetor import convert2prov, decorate, genBindings, praetorTemplates
from praetor.python_builtin_wrap_templateGen import createOpenTemplate

# import convert2prov, decorate, praetorTemplates
# from python_builtin_wrap_templateGen import createOpenTemplate

newDirectories = ['prov', 'json', 'templates', 'big_entities', 'function_store']
for name in newDirectories:
    try:
        subprocess.call(['mkdir', currentDir+name])
    except:
        print('Directory {} already exists'.format(name))

with open(currentDir+'json/'+praetor_settings.json_file,'w') as f:
    f.write("")

if __name__ == '__main__':
    exampleScript = sys.argv[1]

    with open(exampleScript,'r') as f:
        script = f.read()

    newName, commented_globals = decorate.decorateFile(exampleScript, modules=praetor_settings.modules,
                                                       wrap_open=praetor_settings.wrap_open)

    functionCommands = ['python', newName]

    if len(sys.argv) > 2:
        stringCommands = [str(x) for x in sys.argv[2:]]
        functionCommands.extend(stringCommands)

    if praetor_settings.wrap_open:
        createOpenTemplate()

    time_setup = time.time() - t_start
    t_se = time.time()

    with open(currentDir+'/console_output.log', 'wb') as f:
        process = subprocess.Popen(functionCommands, stdout=subprocess.PIPE)
        for c in iter(lambda: process.stdout.read(1), b""):
            sys.stdout.write(c.decode("utf-8"))
            f.write(c)

    example_sep = praetorTemplates.variableSeparation(script)
    tot_sep_dict = {x[0]:[x[1],x[2]] for x in example_sep.values()}
    function_code = [x[3] for x in example_sep.values()]

    for counter,key in enumerate(example_sep.keys()):
        template_no_special = praetorTemplates.escape_forbidden_characters(example_sep[key][0])
        if not os.path.isfile(currentDir+'/templates/{}_template.provn'.format(template_no_special)):
            bundle, bundleStarted = praetorTemplates.generateBundle(example_sep[key], praetorTemplates.bundleStart,
                                                                    praetorTemplates.bundleEnd)
            with open(currentDir+'/templates/{}_template.provn'.format(template_no_special),'w') as f:
                f.write(bundle)

            with open(currentDir+'/templates/{}started_template.provn'.format(template_no_special),'w') as f:
                f.write(bundleStarted)

            with open(currentDir+'/function_store/{}_function.txt'.format(template_no_special),'w') as f:
                f.write(function_code[counter])

    subprocess.call(['mv', newName, currentDir])

    make_string = """
ts := $(shell /bin/date "+%Y%m%d%H%M%S")

build:
\tmvn clean install

merge:
\tfind ./prov/*.provn | sed "s/^/file, /; s/$$/, provn/" | provconvert -merge - -outfile mergey.provn

flatten:
\tprovconvert -infile mergey.provn -index -flatten -outfile flatten_file.provn

flattensvg:
\tprovconvert -infile mergey.provn -index -flatten -outfile flatten_file.svg
  
flattenpng:
\tprovconvert -infile mergey.provn -index -flatten -outfile flatten_file.png

flattenrdf:
\tprovconvert -infile mergey.provn -index -flatten -outfile flatten_file.rdf

flattenttl:
\tprovconvert -infile mergey.provn -index -flatten -outfile flatten_file.ttl    
    """

    t_run = time.time() - t_se
    t_r = time.time()
    if praetor_settings.build_prov:

        convert2prov.combine_prov(currentDir+'json/full_json.json', currentDir+'json/agent_json.json',
                                  currentDir+'templates/', currentDir+'mergey.provn', commented_globals)

        with open(currentDir+'Makefile', 'w') as f:
            f.write(make_string)

        if praetor_settings.flatten_prov:
            os.chdir(currentDir)
            subprocess.call(['make','flatten'])

    t_build = time.time() - t_r

timing_string = "set_up_time = {0}\nrun_time = {1}\nbuild_time = {2}".format(time_setup, t_run, t_build)

with open(currentDir+'prtr_timing_'+pipeline_id+'.txt', 'w') as f:
    f.write(timing_string)



import subprocess

import praetor_settings

if praetor_settings.dynamic_file_dir != '':
    import sys
    sys.path.append(praetor_settings.dynamic_file_dir)
import python_modules


def get_agent_string():
    versions = []
    procPy = subprocess.Popen(['python','--version'],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
    pythonVersion = procPy.stderr.read()
    loc = 'Python '
    pyVer = pythonVersion[pythonVersion.find(loc)+len(loc):]
    pyVer = pyVer.strip()
    pyVer = ''.join(e for e in pyVer if e.isalnum() or e == '.' or e == '-')

    reload(python_modules) # Do I need this ???

    for mod in python_modules.modules:
        proc1 = subprocess.Popen(['pip','show',mod],stdout=subprocess.PIPE)
        proc2 = subprocess.Popen(['grep','Version'],stdin=proc1.stdout,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        proc1.stdout.close()

        out, err = proc2.communicate()
        versions.append(out)

    versions = [x.replace('\n','') for x in versions]
    versions = [x[x.find(': ')+2:] for x in versions]

    modString = ["exe:{}='u2p:{}'".format(x,y) for x,y in zip(python_modules.modules,versions)]
    modString.insert(0,"exe:python='u2p:{}'".format(pyVer))

    agentString = ', '.join(modString)

    return agentString



from setuptools import setup

setup(name='praetor',
      version='2.0',
      python_requires='>=2.7',
      description='Automatic Generation of Provenance from Python Scripts',
      url='https://gitlab.mpcdf.mpg.de/PRAETOR/prov-PRAETOR_public/',
      author='Michael Johnson',
      author_email='michael.johnson0100@gmail.com',
      license='MIT',
      packages=['praetor'],
      install_requires=['codegen','memory_profiler','pymongo','rdflib','pandas'],
      scripts=['bin/praetor_run.py','bin/add_quality.py'],
      zip_safe=False)

